// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO4MF/16AO4MF_Linux_2.x.x.x_GSC_DN/driver/write.c $
// $Rev: 52837 $
// $Date: 2023-04-11 12:37:25 -0500 (Tue, 11 Apr 2023) $

// 16AO4MF: Device Driver: source file

#include "main.h"



//*****************************************************************************
static u32 _buffer_size(chan_data_t* chan)
{
	static const u32	list[16]	=
	{
		4,			8,			16,			32,
		64,			128,		256,		512,
		_1K,		_1K * 2,	_1K * 4,	_1K * 8,
		_1K * 16,	_1K * 32,	0,			0
	};

	u32	bor;
	u32	size;

	bor		= os_reg_mem_rx_u32(chan->dev, chan->vaddr.gsc_bor_32);
	size	= bor & 0xF;
	size	= list[size];
	return(size);
}



//*****************************************************************************
static void _dev_io_sw_init(chan_data_t* chan, dev_io_t* io)
{
	io->io_mode			= AO4MF_IO_MODE_DEFAULT;
	io->timeout_s		= AO4MF_IO_TIMEOUT_DEFAULT;
}



//*****************************************************************************
static void _dev_io_close(chan_data_t* chan, dev_io_t* io)
{
	io->dma_channel	= NULL;
}



//*****************************************************************************
static void _dev_io_open(chan_data_t* chan, dev_io_t* io)
{
	_dev_io_sw_init(chan, io);
}



//*****************************************************************************
static int _dev_io_startup(chan_data_t* chan, dev_io_t* io)
{
	// There is nothing for us to do.
	return(0);
}



//*****************************************************************************
static long _dev_pio_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	u32		bor;
	u32		samples;
	u32		size;

	size	= _buffer_size(chan);
	bor		= os_reg_mem_rx_u32(chan->dev, chan->vaddr.gsc_bor_32);

	if (bor & D15)		// FULL
	{
		samples	= 0;
	}
	else if (bor & D14)	// > 3/4 FULL
	{
		samples	= 1;
	}
	else if (bor & D12)	// EMPTY
	{
		samples	= size;
	}
	else if (bor & D13)	// < 1/4 FULL
	{
		samples	= size * 3 / 4 + 1;
	}
	else				// MEDIUM (1/4 to 3/4 full)
	{
		samples	= size / 4;
	}

	if (samples)
		bytes	= (samples - 1) * 4;	// Data loss is possible if it fills completely.
	else
		bytes	= 0;

	return(bytes);
}



//*****************************************************************************
static long _dev_pio_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	qty	= gsc_write_pio_work_32_bit(chan, io, mem, bytes, st_end);
	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	u32		bor;
	u32		samples;
	u32		size;

	size	= _buffer_size(chan);
	bor		= os_reg_mem_rx_u32(chan->dev, chan->vaddr.gsc_bor_32);

	if (bor & D15)		// FULL
		samples	= 0;
	else if (bor & D14)	// > 3/4 FULL, which leaves too little free space
		samples	= 0;
	else if (bor & D12)	// EMPTY
		samples	= size;
	else if (bor & D13)	// < 1/4 FULL
		samples	= size * 3 / 4 + 1;
	else				// MEDIUM (1/4 to 3/4 full)
		samples	= size / 4;

	if (samples)
		bytes	= (samples - 1) * 4;	// Data loss is possible if it fills completely.
	else
		bytes	= 0;

	return(bytes);
}



//*****************************************************************************
static long _dev_bmdma_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty;
	long			samples	= bytes / 4;
	gsc_dma_setup_t	setup;

	if (samples < io->pio_threshold)
	{
		qty	= gsc_write_pio_work_32_bit(chan, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= chan;
		setup.dev		= chan->dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_BMDMA_WRITE;

		setup.mode		= GSC_DMA_MODE_BLOCK_DMA
						| GSC_DMA_MODE_SIZE_32_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_HOST_TO_BOARD
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



//*****************************************************************************
static long _dev_dmdma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	// The 16AO4MF does not support DMDMA.
	return(-EPERM);
}



//*****************************************************************************
static long _dev_dmdma_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	// The 16AO4MF does not support DMDMA.
	return(-EPERM);
}



/******************************************************************************
*
*	Function:	dev_write_create
*
*	Purpose:
*
*		Perform a one-tine setup of the Write streaming I/O structure.
*
*	Arguments:
*
*		chan	The data for the channel of interest.
*
*		io		The I/O structure for this stream.
*
*	Returned:
*
*		None.
*
******************************************************************************/

int dev_write_create(chan_data_t* chan, dev_io_t* io)
{
	int	ret;

	io->bytes_per_sample	= 4;
	io->io_reg_offset		= 0x18 + (chan->index * 0x40);
	io->io_reg_vaddr		= chan->vaddr.gsc_odbr_32;
	io->pio_threshold		= 32;

	io->dev_io_sw_init		= _dev_io_sw_init;
	io->dev_io_close		= _dev_io_close;
	io->dev_io_open			= _dev_io_open;
	io->dev_io_startup		= _dev_io_startup;
	io->dev_pio_available	= _dev_pio_available;
	io->dev_bmdma_available	= _dev_bmdma_available;
	io->dev_dmdma_available	= _dev_dmdma_available;
	io->dev_pio_xfer		= _dev_pio_xfer;
	io->dev_bmdma_xfer		= _dev_bmdma_xfer;
	io->dev_dmdma_xfer		= _dev_dmdma_xfer;

	io->wait.abort			= AO4MF_WAIT_IO_TX_ABORT;
	io->wait.done			= AO4MF_WAIT_IO_TX_DONE;
	io->wait.error			= AO4MF_WAIT_IO_TX_ERROR;
	io->wait.timeout		= AO4MF_WAIT_IO_TX_TIMEOUT;

	ret	= gsc_io_create(chan, io, _32K);
	return(ret);
}



